//**************************************************************************
//*                     This file is part of the                           *
//*                      Mpxplay - audio player.                           *
//*                  The source code of Mpxplay is                         *
//*        (C) copyright 1998-2005 by PDSoft (Attila Padar)                *
//*                    http://mpxplay.cjb.net                              *
//*                  email: mpxplay@freemail.hu                            *
//**************************************************************************
//*  This program is distributed in the hope that it will be useful,       *
//*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
//*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                  *
//*  Please contact with the author (with me) if you want to use           *
//*  or modify this source.                                                *
//**************************************************************************
//function: ID3v1,ID3v2,APETag read/write

#include "in_file.h"
#include "newfunc\newfunc.h"
#include "playlist\playlist.h"
#include <string.h>

//--------------------------------------------------------------------------
//ID3v1

static char *id3v1genres[]={
  "Blues", "Classic Rock", "Country", "Dance", "Disco", "Funk", "Grunge",
  "Hip-Hop", "Jazz", "Metal", "New Age", "Oldies", "Other", "Pop", "R&B",
  "Rap", "Reggae", "Rock", "Techno", "Industrial", "Alternative", "Ska",
  "Death Metal", "Pranks", "Soundtrack", "Euro-Techno", "Ambient",
  "Trip-Hop", "Vocal", "Jazz+Funk", "Fusion", "Trance", "Classical",
  "Instrumental", "Acid", "House", "Game", "Sound Clip", "Gospel", "Noise",
  "AlternRock", "Bass", "Soul", "Punk", "Space", "Meditative",
  "Instrumental Pop", "Instrumental Rock", "Ethnic", "Gothic", "Darkwave",
  "Techno-Industrial", "Electronic", "Pop-Folk", "Eurodance", "Dream",
  "Southern Rock", "Comedy", "Cult", "Gangsta", "Top 40", "Christian Rap",
  "Pop/Funk", "Jungle", "Native American", "Cabaret", "New Wave",
  "Psychadelic", "Rave", "Showtunes", "Trailer", "Lo-Fi", "Tribal",
  "Acid Punk", "Acid Jazz", "Polka", "Retro", "Musical", "Rock & Roll",
  "Hard Rock", "Folk", "Folk/Rock", "National Folk", "Swing", "Fast-Fusion",
  "Bebob", "Latin", "Revival", "Celtic", "Bluegrass", "Avantgarde",
  "Gothic Rock", "Progressive Rock", "Psychedelic Rock", "Symphonic Rock",
  "Slow Rock", "Big Band", "Chorus", "Easy Listening", "Acoustic", "Humour",
  "Speech", "Chanson", "Opera", "Chamber Music", "Sonata", "Symphony",
  "Booty Bass", "Primus", "Porn Groove", "Satire", "Slow Jam", "Club",
  "Tango", "Samba", "Folklore", "Ballad", "Power Ballad", "Rhythmic Soul",
  "Freestyle", "Duet", "Punk Rock", "Drum Solo", "A capella", "Euro-House",
  "Dance Hall", "Goa", "Drum & Bass", "Club House", "Hardcore", "Terror",
  "Indie", "BritPop", "NegerPunk", "Polsk Punk", "Beat", "Christian Gangsta",
  "Heavy Metal", "Black Metal", "Crossover", "Contemporary C",
  "Christian Rock", "Merengue", "Salsa", "Thrash Metal", "Anime", "JPop",
  "SynthPop"
};

#define MAX_ID3GENRENUM (sizeof(id3v1genres)/sizeof(char *))

static unsigned int id3v10_partlens[5]={30,30,30,4,30};

unsigned int mpxplay_tagging_id3v1_check(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds)
{
 char tag[4];
 if(fbfs->fseek(fbds,-128,SEEK_END)<0)
  return 0;
 if(fbfs->fread(fbds,tag,3)!=3)
  return 0;
 if(tag[0]=='T' && tag[1]=='A' && tag[2]=='G')
  return 1;
 return 0;
}

//always call check_id3tag_v1 before this
char *mpxplay_tagging_id3v1_get(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds,char **id3ip,char *id3p,struct mpxplay_textconv_func_s *mpxplay_textconv_funcs)
{
 unsigned int i,partlen,reallen;
 char *readp,readtmp[128];

 readp=&readtmp[0];
 if(fbfs->fread(fbds,readp,128-3)!=(128-3))
  return id3p;

 for(i=0;i<5;i++){
  partlen=id3v10_partlens[i];
  if(!id3ip[i]){ // maybe we've got it from ID3v2
   pds_strncpy(id3p,readp,partlen);
   if(i==I3I_COMMENT){
    if(id3p[partlen-1]<32) // tracknumber check
     id3p[partlen-1]=0;
   }
   id3p[partlen]=0;
   reallen=pds_strlenc(id3p,32);
   if(reallen){
    id3ip[i]=id3p;
    id3p[reallen]=0;
    reallen=mpxplay_textconv_funcs->all_to_char(id3p,reallen,0);
    id3p+=reallen+1;
   }
  }
  readp+=partlen;
 }
 //tracknumber from id3v1.1
 if(!id3ip[I3I_TRACKNUM]){  // maybe we've got it from ID3v2
  i=(unsigned int)readp[-1];
  if(i && !readp[-2]){
   id3ip[I3I_TRACKNUM]=id3p;
   id3p+=sprintf(id3p,"%d",i)+1;
  }
 }
 //genre
 if(!id3ip[I3I_GENRE]){     // maybe we've got it from ID3v2
  i=(unsigned int)readp[0];
  if(i<MAX_ID3GENRENUM)
   id3ip[I3I_GENRE]=id3v1genres[i];
 }
 return id3p;
}

char *mpxplay_tagging_id3v1_index_to_genre(unsigned int i)
{
 if(i<MAX_ID3GENRENUM)
  return id3v1genres[i];
 return NULL;
}

static unsigned int tag_id3v1_get_genrenum(char *genrename)
{
 unsigned int i;

 if(genrename)
  for(i=0;i<MAX_ID3GENRENUM;i++)
   if(stricmp(id3v1genres[i],genrename)==0)
    return i;

 return 255;
}

int mpxplay_tagging_id3v1_put(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds,struct mpxplay_infile_info_s *miis,char **id3ip,struct mpxplay_textconv_func_s *mpxplay_textconv_funcs)
{
 unsigned int error=MPXPLAY_ERROR_INFILE_WRITETAG_UNKNOWN;
 long fileoffset;
 char strtemp[192];

 if(mpxplay_tagging_id3v1_check(fbfs,fbds))        // is file tagged already?
  fileoffset=fbfs->fseek(fbds,-128,SEEK_END);
 else
  fileoffset=fbfs->fseek(fbds,   0,SEEK_END);
 if(fileoffset>0){                // successfull file positioning
  if(id3ip[I3I_TRACKNUM] && id3ip[I3I_TRACKNUM][0]){       // ID3v1.1
   unsigned int tracknum=pds_atol(id3ip[I3I_TRACKNUM]);
   if(tracknum>255)
    tracknum=255;
   sprintf(strtemp,"TAG%-30.30s%-30.30s%-30.30s%-4.4s%-28.28s%c%c%c",
                   (id3ip[I3I_TITLE])?  id3ip[I3I_TITLE]:"",
                   (id3ip[I3I_ARTIST])? id3ip[I3I_ARTIST]:"",
                   (id3ip[I3I_ALBUM])?  id3ip[I3I_ALBUM]:"",
                   (id3ip[I3I_YEAR])?   id3ip[I3I_YEAR]:"",
                   (id3ip[I3I_COMMENT])?id3ip[I3I_COMMENT]:"",0,
                   tracknum,
                   tag_id3v1_get_genrenum(id3ip[I3I_GENRE]));
  }else{                                                   // ID3v1.0
   sprintf(strtemp,"TAG%-30.30s%-30.30s%-30.30s%-4.4s%-30.30s%c",
                   (id3ip[I3I_TITLE])?  id3ip[I3I_TITLE]:"",
                   (id3ip[I3I_ARTIST])? id3ip[I3I_ARTIST]:"",
                   (id3ip[I3I_ALBUM])?  id3ip[I3I_ALBUM]:"",
                   (id3ip[I3I_YEAR])?   id3ip[I3I_YEAR]:"",
                   (id3ip[I3I_COMMENT])?id3ip[I3I_COMMENT]:"",
                   tag_id3v1_get_genrenum(id3ip[I3I_GENRE]));
  }
  if(fbfs->fwrite(fbds,strtemp,128)==128) // successfull TAG writing
   error=MPXPLAY_ERROR_INFILE_OK;
 }
 return error;
}

//---------------------------------------------------------------------
//ID3v2

#define ID3V2_HEADSIZE 10
#define ID3V2_FOOTERSIZE 10
#define ID3V2_FRAMEHEADSIZE 10
#define ID3V2_MAX_DATALEN 256

static unsigned long mpx_id3v2x_get_framesize_4x7(char *bufp)
{
 unsigned long framesize=0;

 if(!(bufp[1]&0x80))
  framesize+=(bufp[0]&0x7f) << 21;

 if(!(bufp[2]&0x80))
  framesize+=(bufp[1]&0x7f) << 14;

 if(!(bufp[3]&0x80))
  framesize+=(bufp[2]&0x7f) <<  7;

 framesize+=(bufp[3]&0x7f)  <<  0;

 return framesize;
}

static unsigned long mpx_id3v2x_get_framesize_4x8(char *bufp)
{
 return PDS_GETB_BE32(bufp);
}

// needs 10 bytes in bufp !
unsigned long mpxplay_tagging_id3v2_totalsize(char *bufp)
{
 unsigned long id3v2totalsize;
 unsigned int footer_present;

 if(PDS_GETB_LE24(bufp)!=PDS_GET4C_LE32('I','D','3',0))
  return 0;

 bufp+=3;   // ID3
 bufp+=2;   // version

 footer_present=bufp[0]&0x10;
 bufp+=1;   // flags

 id3v2totalsize=ID3V2_HEADSIZE+mpx_id3v2x_get_framesize_4x7(bufp);
 if(footer_present)
  id3v2totalsize+=ID3V2_FOOTERSIZE;

 return id3v2totalsize;
}

//-------------------------------------------------------------------------
static unsigned int mpx_id3v2x_textconv_utf_auto(char *id3p,unsigned int datalen,unsigned int text_enc_type,struct mpxplay_textconv_func_s *mpxplay_textconv_funcs)
{
 if((*(mpxplay_textconv_funcs->control))&ID3TEXTCONV_UTF_AUTO){
  switch(text_enc_type){
   case 1:datalen=mpxplay_textconv_funcs->utf16LE_to_char(id3p,datalen);mpxplay_textconv_funcs->convdone=ID3TEXTCONV_UTF16;break;
   case 2:datalen=mpxplay_textconv_funcs->utf16BE_to_char(id3p,datalen);mpxplay_textconv_funcs->convdone=ID3TEXTCONV_UTF16;break;
   case 3:datalen=mpxplay_textconv_funcs->utf8_to_char(id3p,datalen);mpxplay_textconv_funcs->convdone=ID3TEXTCONV_UTF8;break;
  }
 }
 return datalen;
}

static unsigned int mpx_id3v2x_tagread_text(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds,char *id3p,unsigned int datalen,char **id3ip,struct mpxplay_textconv_func_s *mpxplay_textconv_funcs)
{
 int text_enc_type=0;
 if(datalen<2)
  return 0;
 if(fbfs->fread(fbds,(char *)&text_enc_type,1)!=1)
  return 0;
 datalen--;
 if(fbfs->fread(fbds,id3p,datalen)!=datalen)
  return 0;
 datalen=mpx_id3v2x_textconv_utf_auto(id3p,datalen,text_enc_type,mpxplay_textconv_funcs);
 return datalen;
}

static char *mpx_id3v2x_tagend_text(char *id3p,unsigned int datalen,char **id3ip,struct mpxplay_textconv_func_s *mpxplay_textconv_funcs)
{
 if(datalen){
  id3ip[0]=id3p;
  id3p[datalen]=0;
  datalen=mpxplay_textconv_funcs->all_to_char(id3p,datalen,mpxplay_textconv_funcs->convdone);
  id3p+=datalen+1;
 }
 return id3p;
}

static char *mpx_id3v2x_tagread_txxx(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds,char *id3p,unsigned int datalen,char **id3ip,struct mpxplay_textconv_func_s *mpxplay_textconv_funcs)
{
 datalen=mpx_id3v2x_tagread_text(fbfs,fbds,id3p,datalen,id3ip,mpxplay_textconv_funcs);
 id3p=mpx_id3v2x_tagend_text(id3p,datalen,id3ip,mpxplay_textconv_funcs);
 return id3p;
}

static char *mpx_id3v2x_tagread_tcon(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds,char *id3p,unsigned int datalen,char **id3ip,struct mpxplay_textconv_func_s *mpxplay_textconv_funcs)
{
 datalen=mpx_id3v2x_tagread_text(fbfs,fbds,id3p,datalen,id3ip,mpxplay_textconv_funcs);
 if(datalen){
  if(id3p[0]=='('){
   int gennum=pds_atol(&id3p[1]);
   if((gennum || id3p[1]=='0') && gennum<MAX_ID3GENRENUM)
    id3ip[0]=id3v1genres[gennum];
   else
    id3p=mpx_id3v2x_tagend_text(id3p,datalen,id3ip,mpxplay_textconv_funcs);
  }else{
   id3p=mpx_id3v2x_tagend_text(id3p,datalen,id3ip,mpxplay_textconv_funcs);
  }
 }
 return id3p;
}

static char *mpx_id3v2x_tagread_comm(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds,char *id3p,unsigned int datalen,char **id3ip,struct mpxplay_textconv_func_s *mpxplay_textconv_funcs)
{
 char *s;
 int text_enc_type=0;

 if(datalen<=4)    // real datalen == 0
  return id3p;
 if(fbfs->fread(fbds,(char *)&text_enc_type,1)!=1) // text encoding type (1 byte)
  return id3p;
 if(fbfs->fread(fbds,id3p,3)!=3)                   // language           (3 bytes)
  return id3p;
 datalen-=4;
 if(fbfs->fread(fbds,id3p,datalen)!=datalen)
  return id3p;

 datalen=mpx_id3v2x_textconv_utf_auto(id3p,datalen,text_enc_type,mpxplay_textconv_funcs);

 s=id3p;
 do{
  if((datalen<1) || *s)
   break;
  datalen--;
  s++;
 }while(1);

 if(datalen && (s>id3p))
  pds_memcpy(id3p,s,datalen);

 id3p=mpx_id3v2x_tagend_text(id3p,datalen,id3ip,mpxplay_textconv_funcs);

 return id3p;
}

typedef struct id3v2x_one_frame_info_s{
 char *frameid;
 char *(*frame_reader)(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds,char *id3p,unsigned int datalen,char **id3ip,struct mpxplay_textconv_func_s *mpxplay_textconv_funcs);
 unsigned int i3i_index;
}id3v2x_one_frame_info_s;

typedef struct id3v2x_one_version_info_s{
 unsigned char version_number;
 unsigned long (*frame_getsize)(char *bufp);
 id3v2x_one_frame_info_s *(*frame_selector)(char *framehead,id3v2x_one_frame_info_s *f0);
 id3v2x_one_frame_info_s *supported_frames;
}id3v2x_one_version_info_s;

static id3v2x_one_frame_info_s id3v23_supported_frames[]=
{{"TPE1",&mpx_id3v2x_tagread_txxx,I3I_ARTIST},
 {"TIT2",&mpx_id3v2x_tagread_txxx,I3I_TITLE},
 {"TALB",&mpx_id3v2x_tagread_txxx,I3I_ALBUM},
 {"TYER",&mpx_id3v2x_tagread_txxx,I3I_YEAR},
 {"TCON",&mpx_id3v2x_tagread_tcon,I3I_GENRE},
 {"COMM",&mpx_id3v2x_tagread_comm,I3I_COMMENT},
 {"TRCK",&mpx_id3v2x_tagread_txxx,I3I_TRACKNUM},
 {NULL,NULL,0}
};

static id3v2x_one_frame_info_s id3v24_supported_frames[]=
{{"TPE1",&mpx_id3v2x_tagread_txxx,I3I_ARTIST},
 {"TIT2",&mpx_id3v2x_tagread_txxx,I3I_TITLE},
 {"TALB",&mpx_id3v2x_tagread_txxx,I3I_ALBUM},
 {"TYER",&mpx_id3v2x_tagread_txxx,I3I_YEAR},
 {"TDRC",&mpx_id3v2x_tagread_txxx,I3I_YEAR},
 {"TCON",&mpx_id3v2x_tagread_tcon,I3I_GENRE},
 {"COMM",&mpx_id3v2x_tagread_comm,I3I_COMMENT},
 {"TRCK",&mpx_id3v2x_tagread_txxx,I3I_TRACKNUM},
 {NULL,NULL,0}
};

static id3v2x_one_frame_info_s *mpx_id3v2x_frame_selector(char *framehead,id3v2x_one_frame_info_s *f)
{
 while(f->frameid){
  if(pds_strncmp(framehead,f->frameid,4)==0)
   return f;
  f++;
 }
 return NULL;
}

//2.3 and 2.4 are supported
static id3v2x_one_version_info_s id3v2_all_version_infos[]=
{{0x03,mpx_id3v2x_get_framesize_4x8,mpx_id3v2x_frame_selector,&id3v23_supported_frames[0]},
 {0x04,mpx_id3v2x_get_framesize_4x7,mpx_id3v2x_frame_selector,&id3v24_supported_frames[0]},
 {0x00,NULL,NULL}
};

static id3v2x_one_version_info_s *mpx_id3v2_version_selector(char *verstr)
{
 id3v2x_one_version_info_s *v=&id3v2_all_version_infos[0];
 while(v->frame_getsize){
  if((unsigned char)verstr[0]==v->version_number)
   return v;
  v++;
 }
 return NULL;
}

char *mpxplay_tagging_id3v2_get(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds,struct mpxplay_infile_info_s *miis,char **id3ip,char *id3p,struct mpxplay_textconv_func_s *mpxplay_textconv_funcs)
{
 id3v2x_one_version_info_s *version_handler;
 unsigned long id3v2totalsize,filepos;
 unsigned int extended_header,footer_present;
 //unsigned int unsynchronisation,experimental_indicator;
 char *bufp;
 char readtmp[16];

 if(fbfs->fseek(fbds,0,SEEK_SET)<0)
  return id3p;
 if(fbfs->fread(fbds,&readtmp[0],ID3V2_HEADSIZE)!=ID3V2_HEADSIZE)
  return id3p;
 bufp=&readtmp[0];

 //"ID3"
 if(PDS_GETB_LE24(bufp)!=PDS_GET4C_LE32('I','D','3',0))
  return id3p;
 bufp+=3;

 //version
 version_handler=mpx_id3v2_version_selector(bufp);
 if(!version_handler)
  return id3p;
 bufp+=2;

 //flags
 //unsynchronisation     =bufp[0]&0x80;
 extended_header       =bufp[0]&0x40;
 //experimental_indicator=bufp[0]&0x20;
 footer_present        =bufp[0]&0x10;
 bufp+=1;

 id3v2totalsize=ID3V2_HEADSIZE+mpx_id3v2x_get_framesize_4x7(bufp);
 if(footer_present)
  id3v2totalsize+=ID3V2_FOOTERSIZE;

 //fprintf(stderr,"id3v2size:%d %4.4X \n",id3v2totalsize,id3v2totalsize);

 if(id3v2totalsize>=fbfs->filelength(fbds) || id3v2totalsize<=ID3V2_HEADSIZE)
  return id3p;

 filepos=ID3V2_HEADSIZE;

 if(extended_header){
  unsigned long ehsize;
  if(fbfs->fread(fbds,&readtmp[0],4)!=4)
   return id3p;
  ehsize=version_handler->frame_getsize(&readtmp[0]);
  if(ehsize<4 || (id3v2totalsize+ehsize)>=fbfs->filelength(fbds))
   return id3p;
  filepos+=ehsize;
  if(fbfs->fseek(fbds,ehsize-4,SEEK_CUR)<0)
   return id3p;
 }

 if((filepos+ID3V2_FRAMEHEADSIZE)>=id3v2totalsize){ // no frames (tags) in the ID3V2
  fbfs->fseek(fbds,id3v2totalsize,SEEK_SET);
  return id3p;
 }

 do{
  unsigned int datalen;
  id3v2x_one_frame_info_s *frameselect;
  // maybe there are padding bytes... required ???
  /*do{
   fbfs->fread(fbds,&readtmp[0],1);
   filepos++;
   if(filepos>=id3v2totalsize)
    return id3p;
  }while(!readtmp[0]);
  filepos--;
  fbfs->fread(fbds,&readtmp[1],ID3V2_FRAMEHEADSIZE-1);*/

  if(fbfs->fread(fbds,&readtmp[0],ID3V2_FRAMEHEADSIZE)!=ID3V2_FRAMEHEADSIZE)
   break;
  filepos+=ID3V2_FRAMEHEADSIZE;

  datalen=version_handler->frame_getsize(&readtmp[4]);

  //fprintf(stderr,"\n%c%c%c%c %2d : ",readtmp[0],readtmp[1],readtmp[2],readtmp[3],datalen);

  if(datalen){
   filepos+=datalen;
   if(filepos>id3v2totalsize)
    break;
   if((datalen<ID3V2_MAX_DATALEN) && ((frameselect=version_handler->frame_selector(&readtmp[0],version_handler->supported_frames))!=NULL)){
    unsigned int i3i_select=frameselect->i3i_index;

    if(!id3ip[i3i_select]){
     //fprintf(stderr,"%s : ",frameselect->frameid);
     mpxplay_textconv_funcs->convdone=0;
     id3p=frameselect->frame_reader(fbfs,fbds,id3p,datalen,&id3ip[i3i_select],mpxplay_textconv_funcs);
     //fprintf(stderr," %s ",((id3ip[i3i_select])? id3ip[i3i_select]:"none"));
     if(id3ip[i3i_select])
      continue;
    }
   }
   if(fbfs->fseek(fbds,filepos,SEEK_SET)<0)
    break;
  }
 }while(filepos<(id3v2totalsize-ID3V2_FRAMEHEADSIZE));

 if(filepos!=id3v2totalsize)
  fbfs->fseek(fbds,id3v2totalsize,SEEK_SET);

 return id3p;
}

//--------------------------------------------------------------------
//APEtag

typedef struct APETag_s {
 unsigned char   ID       [8];    // should equal 'APETAGEX'
 unsigned char   Version  [4];    // 1000 or 2000
 unsigned char   Length   [4];    // the complete size of the tag, including this footer
 unsigned char   TagCount [4];    // the number of fields in the tag
 unsigned char   Flags    [4];    // the tag flags (none currently defined)
 unsigned char   Reserved [8];    // reserved for later use
}APETag_s;

typedef struct tagtype_s{
 char *name;
 unsigned int i3i_index;
}tagtype_s;

static tagtype_s tagtypes[]={
 {"Artist",I3I_ARTIST},{"Title"  ,I3I_TITLE}  ,{"Album",I3I_ALBUM},
 {"Year"  ,I3I_YEAR}  ,{"Comment",I3I_COMMENT},{"Genre",I3I_GENRE},
 {"Track" ,I3I_TRACKNUM}
};

#define TAGTYPENUM (sizeof(tagtypes)/sizeof(struct tagtype_s))

static int ape_tag_check_apetag(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds,struct APETag_s *apetag)
{
 int version,taglen;
 long filesize=fbfs->filelength(fbds);
 if(filesize<=((long)sizeof(struct APETag_s)))
  return 0;
 if(fbfs->fseek(fbds,-((long)sizeof(struct APETag_s)),SEEK_END)<0)
  return 0;
 if(fbfs->fread(fbds,(char *)apetag,sizeof(struct APETag_s))!=sizeof(struct APETag_s))
  return 0;
 if(pds_strncmp(apetag->ID,"APETAGEX",sizeof(apetag->ID))!=0)
  return 0;
 version=PDS_GETB_LE32(apetag->Version);
 if(version!=1000 && version!=2000)
  return 0;
 taglen=PDS_GETB_LE32(apetag->Length);
 if(taglen<=sizeof(struct APETag_s) || (taglen>=filesize))
  return 0;
 if(!PDS_GETB_LE32(apetag->TagCount))
  return 0;
 if(fbfs->fseek(fbds,-taglen,SEEK_END)<0)
  return 0;
 return 1;
}

int mpxplay_tagging_apetag_check(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds)
{
 struct APETag_s apetag;
 return ape_tag_check_apetag(fbfs,fbds,&apetag);
}

static char *ape_tag_read_apetag(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds,struct APETag_s *apetag,struct mpxplay_infile_info_s *miis,char **id3ip,char *id3p,struct mpxplay_textconv_func_s *mpxplay_textconv_funcs)
{
 long tt,version,taglen,tagcount;
 long namelen,datalen,dataflags;
 char *tagbuff,*tbp,*tname,*tdata;

 taglen=PDS_GETB_LE32(apetag->Length);
 tagbuff=malloc(taglen);
 if(!tagbuff)
  return id3p;
 if(fbfs->fread(fbds,tagbuff,taglen-sizeof(struct APETag_s))!=(taglen-sizeof(struct APETag_s)))
  goto err_out_tagread;
 tbp=tagbuff;
 tagbuff[taglen-sizeof(struct APETag_s)]=0;
 version=PDS_GETB_LE32(apetag->Version);
 tagcount=PDS_GETB_LE32(apetag->TagCount);
 for(;*tbp && tagcount--;){
  datalen  =PDS_GETB_LE32(tbp); tbp+=4;
  dataflags=PDS_GETB_LE32(tbp); tbp+=4;
  namelen=strlen(tbp)+1;
  if(namelen>128)        // impossible data
   break;
  tname=tbp; tbp+=namelen;
  tdata=tbp; tbp+=datalen;
  if(tbp>(&tagbuff[taglen-sizeof(struct APETag_s)])) // bad data -> overflow
   break;
  //fprintf(stdout,"%s %s \n",tname,tdata);
  if((namelen>1) && ((version==1000 && datalen>1) || (version==2000 && datalen>0 && !(dataflags&2)))){
   /*if(pds_stricmp(tname,"replaygain_track_gain")==0){// || pds_stricmp(tname,"replaygain_album_gain")==0)
    miis->audio_decoder_infos->replaygain=atof(tdata);
    //fprintf(stdout,"%s %s %2.2f df:%d\n",tname,tdata,miis->audio_decoder_infos->replaygain,dataflags);
   }else*/{
    for(tt=0;tt<TAGTYPENUM;tt++){
     int i3iindex=tagtypes[tt].i3i_index;
     if(!id3ip[i3iindex] && pds_stricmp(tname,tagtypes[tt].name)==0){
      id3ip[i3iindex]=id3p;
      pds_memcpy(id3p,tdata,datalen);
      if(version==1000){
       id3p+=mpxplay_textconv_funcs->all_to_char(id3p,datalen-1,0)+1; // datalen includes the terminating null in v1
      }else{
       id3p[datalen]=0;
       if((*(mpxplay_textconv_funcs->control))&ID3TEXTCONV_UTF_AUTO)
        datalen=mpxplay_textconv_funcs->utf8_to_char(id3p,datalen);  // allways in v2
       id3p+=mpxplay_textconv_funcs->all_to_char(id3p,datalen,ID3TEXTCONV_UTF8)+1;   // datalen does not include the terminating null in v2
      }
      break;
     }
    }
   }
  }
 }
err_out_tagread:
 free(tagbuff);
 return id3p;
}

char *mpxplay_tagging_apetag_get(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds,struct mpxplay_infile_info_s *miis,char **id3ip,char *id3p,struct mpxplay_textconv_func_s *mpxplay_textconv_funcs)
{
 struct APETag_s apetag;

 if(mpxplay_tagging_id3v1_check(fbfs,fbds))
  return mpxplay_tagging_id3v1_get(fbfs,fbds,id3ip,id3p,mpxplay_textconv_funcs);

 if(ape_tag_check_apetag(fbfs,fbds,&apetag))
  return ape_tag_read_apetag(fbfs,fbds,&apetag,miis,id3ip,id3p,mpxplay_textconv_funcs);

 return id3p;
}

static unsigned int apetag_convert_out(char **extradatas,char *id3p,struct mpxplay_textconv_func_s *mtf,unsigned int do_utf8)
{
 char *ed,*tmp,*dest;
 unsigned int len;

 len=pds_strlen(id3p);
 if(!len)
  return 0;
 len++; // we count the terminating null too

 if(do_utf8){
  ed=(char *)malloc(len*3); // *3 = max utf8 expansion
  if(!ed)
   return 0;
  *extradatas=ed;
  tmp=(char *)malloc(len);
  if(!tmp)
   return 0;
  dest=mtf->char_to_all(tmp,id3p);
  len=mtf->char_to_utf8(ed,dest,len*3);
  free(tmp);
 }else{
  ed=(char *)malloc(len);
  if(!ed)
   return 0;
  *extradatas=ed;
  dest=mtf->char_to_all(ed,id3p);
  if(dest!=ed)
   pds_strcpy(ed,dest);
 }

 return len;
}

int mpxplay_tagging_apetag_put(struct mpxplay_filehand_buffered_func_s *fbfs,void *fbds,struct mpxplay_infile_info_s *miis,char **id3ip,struct mpxplay_textconv_func_s *mpxplay_textconv_funcs)
{
 int tt,i,extratags=0,cutfile=0;
 long alltagsize=0,tagcount=0,newalltagsize,*tagdlens=NULL,*tagflags=NULL;
 mpxp_int32_t tagnamelen,tagdatalen,flags=0,version;
 char *tagbuff=NULL,*tbp,**tagnames=NULL,**tagdatas=NULL,*extradatas[TAGTYPENUM];
 struct APETag_s apetag;

 if(mpxplay_tagging_id3v1_check(fbfs,fbds))
  return mpxplay_tagging_id3v1_put(fbfs,fbds,miis,id3ip,mpxplay_textconv_funcs);

 if(ape_tag_check_apetag(fbfs,fbds,&apetag)){ // APE file already has APETAGEX
  // load all tagdatas, update and write back
  version=PDS_GETB_LE32(apetag.Version);
  if((version!=1000) && (version!=2000))
   return MPXPLAY_ERROR_INFILE_WRITETAG_TAGTYPE;
  alltagsize=PDS_GETB_LE32(apetag.Length);
  tbp=tagbuff=malloc(alltagsize);
  tagcount=PDS_GETB_LE32(apetag.TagCount);
  tagnames=(char **)malloc((tagcount+TAGTYPENUM)*sizeof(char *));
  tagdatas=(char **)malloc((tagcount+TAGTYPENUM)*sizeof(char *));
  tagdlens=(long *)malloc((tagcount+TAGTYPENUM)*sizeof(long));
  tagflags=(long *)malloc((tagcount+TAGTYPENUM)*sizeof(long));
  if(!tagbuff || !tagnames || !tagdatas || !tagdlens || !tagflags)
   goto err_out_tagput;
  pds_memset(extradatas,0,sizeof(extradatas));
  //load
  if(fbfs->fread(fbds,tagbuff,alltagsize-sizeof(struct APETag_s))!=(alltagsize-sizeof(struct APETag_s)))
   goto err_out_tagput;
  for(i=0;i<tagcount && (tbp<(tagbuff+alltagsize-sizeof(struct APETag_s)));i++){
   tagdlens[i]=PDS_GETB_LE32(tbp);tbp+=4;
   tagflags[i]=PDS_GETB_LE32(tbp);tbp+=4;
   tagnames[i]=tbp;               tbp+=strlen(tbp)+1;
   tagdatas[i]=tbp;               tbp+=tagdlens[i];
  }
  tagcount=i; // correction in the case of a wrong ape-tag
  //update
  for(tt=0;tt<TAGTYPENUM;tt++){
   int i3iindex=tagtypes[tt].i3i_index;
   if(id3ip[i3iindex]){
    for(i=0;i<tagcount;i++){
     if((version==1000 && tagdlens[i]>1) || (version==2000 && tagdlens[i]>0 && !(tagflags[i]&2))){
      if(pds_stricmp(tagnames[i],tagtypes[tt].name)==0){ // found, update
       unsigned int len=apetag_convert_out(&extradatas[tt],id3ip[i3iindex],mpxplay_textconv_funcs,((version==2000)? 1:0));
       if(len){
        tagdatas[i]=extradatas[tt];
        tagdlens[i]=len-((version==2000)? 1:0); // v2 does not count terminating null
       }
       break;
      }
     }
    }
    if(i==tagcount){  // not found, new
     unsigned int len=apetag_convert_out(&extradatas[tt],id3ip[i3iindex],mpxplay_textconv_funcs,((version==2000)? 1:0));
     if(len){
      tagnames[tagcount+extratags]=tagtypes[tt].name;
      tagdatas[tagcount+extratags]=extradatas[tt];
      tagdlens[tagcount+extratags]=len-((version==2000)? 1:0);
      tagflags[tagcount+extratags]=0;
      extratags++;
     }
    }
   }
  }
  //write back
  if(fbfs->fseek(fbds,-alltagsize,SEEK_END)<0)
   goto err_out_tagput;
  newalltagsize=0;
  tagcount+=extratags;
  for(i=0;i<tagcount;i++){
   tagnamelen=strlen(tagnames[i])+1;
   fbfs->fwrite(fbds,&tagdlens[i],sizeof(tagdlens[i]));
   fbfs->fwrite(fbds,&tagflags[i],sizeof(tagflags[i]));
   fbfs->fwrite(fbds,tagnames[i],tagnamelen);
   fbfs->fwrite(fbds,tagdatas[i],tagdlens[i]);
   newalltagsize+=sizeof(tagdlens[i])+sizeof(tagflags[i])+tagnamelen+tagdlens[i];
  }
  if(newalltagsize<alltagsize)
   cutfile=1;
  alltagsize=newalltagsize;
  free(tagflags);
  free(tagdlens);
  free(tagdatas);
  free(tagnames);
  free(tagbuff);
  for(tt=0;tt<TAGTYPENUM;tt++){
   if(extradatas[tt])
    free(extradatas[tt]);
  }
 }else{
  // write an absolute new APEv2 tag
  version=2000;
  if(fbfs->fseek(fbds,0,SEEK_END)<0)
   return MPXPLAY_ERROR_INFILE_EOF;
  for(tt=0;tt<TAGTYPENUM;tt++){
   int i3iindex=tagtypes[tt].i3i_index;
   if(id3ip[i3iindex]){
    unsigned int len=apetag_convert_out(&extradatas[0],id3ip[i3iindex],mpxplay_textconv_funcs,((version==2000)? 1:0));
    if(len){
     tagnamelen=strlen(tagtypes[tt].name)+1;
     tagdatalen=len-((version==2000)? 1:0);
     fbfs->fwrite(fbds,&tagdatalen,sizeof(tagdatalen));
     fbfs->fwrite(fbds,&flags,sizeof(flags));
     fbfs->fwrite(fbds,tagtypes[tt].name,tagnamelen);
     fbfs->fwrite(fbds,extradatas[0],tagdatalen);
     alltagsize+=sizeof(tagdatalen)+sizeof(flags)+tagnamelen+tagdatalen;
     tagcount++;
    }
   }
  }
 }
 //finish/close the tag
 memcpy(apetag.ID,"APETAGEX",sizeof("APETAGEX"));
 PDS_PUTB_LE32(apetag.Version,version);
 PDS_PUTB_LE32(apetag.Length,alltagsize+sizeof(struct APETag_s));
 PDS_PUTB_LE32(apetag.TagCount,tagcount);
 PDS_PUTB_LE32(apetag.Flags,0);
 memset(apetag.Reserved,0,sizeof(apetag.Reserved));
 fbfs->fwrite(fbds,(void *)&apetag,sizeof(struct APETag_s));

 if(cutfile)
  fbfs->chsize(fbds,fbfs->ftell(fbds));

 return MPXPLAY_ERROR_INFILE_OK;

err_out_tagput:
 if(tagflags)
  free(tagflags);
 if(tagdlens)
  free(tagdlens);
 if(tagdatas)
  free(tagdatas);
 if(tagnames)
  free(tagnames);
 if(tagbuff)
  free(tagbuff);
 return MPXPLAY_ERROR_INFILE_MEMORY;
}
